﻿using System;
using System.Collections.Generic;
using System.IO;
using TSLab.DataSource;
using TSLab.Script;
using TSLab.Script.Realtime;

namespace FinInfoCache.Common
{
    public class Consts
    {
        public const string Category = "FinInfo";
        public const string DefPath = @"C:\\Data";
        public const string PathName = "Папка";
        public const char Separator = ',';      // Разделитель полей в текстовом файле.
    }

    public static class FininfoCache
    {
        /// <summary>
        /// Формат файла данных. Если добавляются или удаляются данные для записи. Нужно увеличивать число.
        /// </summary>
        public const byte FileFormatVersion = 2;
        /// <summary>
        /// Расширение для файла данных.
        /// </summary>
        public const string Extension = ".fid";
        /// <summary>
        /// Размер заголовка файла. Один байт. Обозначает версию формата файла. 1, 2, 3 и тд. 
        /// </summary>
        public const int HeaderSize = sizeof(byte);
        /// <summary>
        /// Размер записи. Вычисляется как размер всех полей входящих в запись.
        /// </summary>
        public const int RecSize = sizeof(long) + 9 * sizeof(double);
        /// <summary>
        /// Специальный кэш данных. Содержит уникальный ключ, и список записей для этого ключа.
        /// </summary>
        static Dictionary<string, IList<FinInfoBlock>> _dict = new Dictionary<string, IList<FinInfoBlock>>(4);

        #region Запись данных

        /// <summary>
        /// Пишет в конец файла текущие данные по инструменту. Если текущие данные такие же как 
        /// последняя запись в файле, то ничего не пишется.
        /// </summary>
        /// <param name="security"></param>
        /// <param name="path"></param>
        public static void WriteFininfo(ISecurity security, string path)
        {
            // Писать будем только в реалтайм.
            var srt = security as ISecurityRt;
            if (srt == null)
                return;

            // Получаем путь к файлу с данными. Собираем из имени инструмента.
            path = Path.GetFullPath(path);
            if (!Directory.Exists(path))
                Directory.CreateDirectory(path);
            var fname = Path.Combine(path, srt.Symbol + Extension);


            using (var fs = File.Open(fname, FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.Read))
            {
                // Если размер файла больше 0 то проверим его формат. Иначе пишем в начало номер формата.
                if (fs.Length > 0)
                    CheckFormat(fs);
                else
                    fs.WriteByte(FileFormatVersion); // Номер версии формата


                // Читаем последнюю запись файла. Если она есть.
                var recCount = (fs.Length - HeaderSize) / RecSize;
                var lastRecord = new FinInfoBlock();
                if (recCount > 0)
                    lastRecord.ReadFromStream(recCount - 1, new BinaryReader(fs));


                // Если данные последней записи отличаются от текущих данных, то пишем в конец файла текущие данные.
                var fi = srt.FinInfo;
                if (fi != null && FinInfoContainsData(fi) && !lastRecord.EqualsExceptLastUpdate(fi))
                {
                    var pos = HeaderSize + recCount * RecSize;
                    fs.Seek(pos, SeekOrigin.Begin);
                    using (var bw = new BinaryWriter(fs))
                        fi.WriteToStream(bw);
                }
            }
        }

        #endregion

        #region Чтение данных

        public static IList<FinInfoBlock> GetFininfo(ISecurity security, string path)
        {
            var key = GetKey(security);


            IList<FinInfoBlock> list;
            lock (_dict)
            {
                // Если в нашем словаре уже есть данные для текущего уникального ключа, просто вернем их. Не будем читать файл.
                if (_dict.TryGetValue(key, out list))
                    return list;

                list = ReadFininfo(security, path);

                // Если есть ключи для такого же инструмента то удалим их все, независимо от таймфрейма итд.
                // Добавим запись только для текущего инструмента и ключа.
                var keys = new List<string>(_dict.Keys);
                var s = security.Symbol + '.';
                foreach (var k in keys)
                    if (k.StartsWith(s))
                        _dict.Remove(k);

                _dict[key] = list;
            }
            return list;
        }

        /// <summary>
        /// Для каждого бара инструмента, считываем ближайшую запись справа (от времени закрытия бара) в файле.
        /// Если бар имеет дату меньше чем первая запись, то вернется первая запись, если больше последней записи, то последняя запись.
        /// </summary>
        /// <param name="security"></param>
        /// <param name="path"></param>
        /// <returns></returns>
        public static IList<FinInfoBlock> ReadFininfo(ISecurity security, string path)
        {
            var bars = security.Bars;
            int bCount = bars.Count;
            if (bCount <= 0)
                return new FinInfoBlock[0];

            // Получаем имя файла с данными. Оно зависит от инструмента.
            path = Path.GetFullPath(path);
            var fname = Path.Combine(path, security.Symbol + Extension);

            using (var fs = File.Open(fname, FileMode.Open, FileAccess.Read, FileShare.ReadWrite))
            {
                // Проверяем формат файла.
                CheckFormat(fs);

                // Создаем массив для данных, забиваем его ссылкой на один пустой объект.
                var values = new FinInfoBlock[bCount];
                var zeroFi = new FinInfoBlock();
                for (int i = 0; i < bCount; i++)
                    values[i] = zeroFi;

                // Если число записей в файле 0, то вернем массив пустых объектов.
                long recCount = (fs.Length - HeaderSize) / RecSize;
                if (recCount <= 0)
                    return values;

                // Находим время окончания самого первого бара в инструменте.
                // Для этого времени будет искаться первая запись в файле данных.
                // Время приводим к универсальному. Данные все в универсальном времени.
                var startDateUtc = MscToUtc(bars[0].Date).Ticks;
                var tf = GetTimeFrame(security.IntervalBase, security.Interval);
                startDateUtc += tf.Ticks;  // Конец бара

                // Определим номер записи для первого бара инструмента
                var startRec = 0L;
                var finishRec = recCount - 1;
                var br = new BinaryReader(fs);
                var curRec = FindStartingRecord(br, startDateUtc, startRec, finishRec);
                if (curRec < 0)
                    return values;

                // Читаем первую и вторую запись. Если вторую считать нельзя, создаем фиктивную с временем очень большим.
                var fb = new FinInfoBlock();
                fb.ReadFromStream(curRec, br);

                var fbNext = new FinInfoBlock { LastUpdate = DateTime.MaxValue };
                if (curRec < finishRec)
                    fbNext.ReadFromStream(curRec + 1, br);


                // Переходим к первому бару инструмента, для которого время окончания больше найденной записи в файле.
                // Запись искали ближайщую справа по времени.
                int starti = 0;
                for (; starti < bCount; starti++)
                {
                    var barTimeUtc = MscToUtc(bars[starti].Date);
                    if (barTimeUtc + tf >= fb.LastUpdate)
                        break;
                }
                // Ищем для каждого бара ближайшую справа запись по времени.
                // Пока не кончатся все записи.
                for (int i = starti; i < bCount; i++)
                {
                    var barCloseTimeUtc = MscToUtc(bars[i].Date + tf);
                    while (barCloseTimeUtc > fbNext.LastUpdate && curRec < finishRec)
                    {
                        fb = fbNext;
                        fbNext = FinInfoCacheHelper.ReadFromStream(curRec++, br);
                    }
                    values[i] = fb;
                }
                return values;
            }
        }

        /// <summary>
        /// Ищем номер записи, ближайший слева к времени date. Если date лежит между 4 и 5 записью, То вернется 4.
        /// Если date меньше чем самая ранняя запись, то вернем 0. Если больше чем самая поздняя то вернем самую последнюю запись.
        /// </summary>
        /// <param name="br"></param>
        /// <param name="date">Дата, в виде числа тиков. Использовать нужно Utc время.</param>
        /// <param name="startRec"></param>
        /// <param name="finishRec"></param>
        /// <returns></returns>
        static long FindStartingRecord(BinaryReader br, long date, long startRec, long finishRec)
        {
            // Если время первой считанной записи больше чем искомое время, вернем первую запись как стартовую.
            var fs = br.BaseStream;
            fs.Seek(HeaderSize + startRec * RecSize, SeekOrigin.Begin);
            var currDate = br.ReadInt64();
            if (date < currDate)
                return startRec;

            // если время последней записи в файле меньше искомого времени то вернем номер последней записи.
            fs.Seek(HeaderSize + finishRec * RecSize, SeekOrigin.Begin);
            currDate = br.ReadInt64();
            if (currDate < date)
                return finishRec;

            // В остальных случаях ищем запись методом деления пополам.
            // Возвращается всегда запись, в которую попадает время date. Если время попало между 4 и 5 записью, То вернется 4 запись.
            long rec = (int)((finishRec + startRec) * 0.66);  // Первая итерация на начало последних 33%
            do
            {
                fs.Seek(HeaderSize + rec * RecSize, SeekOrigin.Begin);
                currDate = br.ReadInt64();
                if (currDate < date)
                    startRec = rec;
                else if (currDate > date)
                    finishRec = rec;
                else
                    return rec;
                rec = (finishRec + startRec) / 2;
            } while (startRec + 1 < finishRec);
            return startRec;
        }

        #endregion

        #region Helper

        /// <summary>
        /// Формируем уникальный ключ из инструмента и его данных.
        /// </summary>
        /// <param name="security">Информация по инструменту.</param>
        /// <returns>Ключ = "Symbol"."5m"."bars[0].Date.Ticks"."bars.last().Date.Ticks"
        /// Используем имя инструмента, таймфрейм, дату начала и дату конца.</returns>
        static string GetKey(ISecurity security)
        {
            string tf = GetTimeFrameCode(security.IntervalBase, security.Interval);
            string key = security.Symbol + '.' + tf;
            var bars = security.Bars;
            if (bars.Count > 0)
                key += '.' + bars[0].Date.Ticks + '.' + bars[bars.Count - 1].Date.Ticks;
            return key;
        }

        /// <summary>
        /// Формирует строку обозначающую интервал. 5m, 10m, 5s, итд. 
        /// Используются стандартные ТСЛаб интервалы d, m, s, t
        /// </summary>
        /// <param name="intervalBase"></param>
        /// <param name="interval"></param>
        /// <returns>Строка интервала.</returns>
        static string GetTimeFrameCode(DataIntervals intervalBase, int interval)
        {
            string tf = interval.ToString();
            switch (intervalBase)
            {
                case DataIntervals.DAYS:
                    tf += 'd';
                    break;
                case DataIntervals.MINUTE:
                    tf += 'm';
                    break;
                case DataIntervals.SECONDS:
                    tf += 's';
                    break;
                case DataIntervals.TICK:
                default:
                    tf += 't';
                    break;
            }
            return tf;
        }

        static TimeSpan GetTimeFrame(DataIntervals intervalBase, int interval)
        {
            TimeSpan tf;
            switch (intervalBase)
            {
                case DataIntervals.DAYS:
                    tf = new TimeSpan(interval, 0, 0, 0);
                    break;
                case DataIntervals.MINUTE:
                    tf = new TimeSpan(0, 0, interval, 0);
                    break;
                case DataIntervals.SECONDS:
                    tf = new TimeSpan(0, 0, 0, interval);
                    break;
                case DataIntervals.TICK:
                default:
                    tf = new TimeSpan(0, 0, 0, 0, interval);
                    break;
            }
            return tf;
        }

        static bool FinInfoContainsData(FinInfo fi)
        {
            return 0 != (fi.Ask ?? 0) ||
                   0 != (fi.Bid ?? 0) ||
                   0 != (fi.BuyCount ?? 0) ||
                   0 != (fi.BuySqty ?? 0) ||
                   0 != (fi.SellCount ?? 0) ||
                   0 != (fi.SellSqty ?? 0) ||
                   0 != (fi.TheoriticPrice ?? 0) ||
                   0 != (fi.Volatility ?? 0) ||
                   0 != (fi.OpenInterest ?? 0);
        }

        /// <summary>
        /// Конвертирует московское время баров, в универсальное время. 
        /// </summary>
        /// <param name="time">Время московское, поле kind == unspecified</param>
        /// <returns></returns>
        static DateTime MscToUtc(DateTime time)
        {
            if (time.Kind != DateTimeKind.Unspecified)
                throw new ArgumentException("Время не должно быть локальным, или UTC временем. Kind == unspecified.");


            var mscTime = time;
            var utcTime = TimeZoneInfo.ConvertTimeToUtc(mscTime, TimeZoneInfo.FindSystemTimeZoneById("Russian Standard Time"));
            return utcTime;
        }

        /// <summary>
        /// Считывает формат файла из потока. Проверяет его на корректность. Генерит исключения если кривой формат.
        /// </summary>
        /// <param name="fs">Поток</param>
        /// <returns>Истина если формат верный, иначе ложь</returns>
        public static bool CheckFormat(FileStream fs)
        {
            if (fs.Length == 0)
                throw new InvalidOperationException("Невозможно считать формат в пустом файле: " + fs.Name);

            fs.Seek(0, SeekOrigin.Begin);
            var formatByte = fs.ReadByte();

            if (formatByte != FileFormatVersion)
                throw new InvalidDataException("Неподходящий формат данных в файле: " + fs.Name);

            return formatByte == FileFormatVersion;
        }

        #endregion

    }
}