﻿using System;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Threading;
using FinInfoCache.Common;
using TSLab.DataSource;

namespace FinInfoCache.Converter
{
    class Program
    {
        // Вместо знаков {0} далее будет подставлен текущий разделитель.
        private const string HeaderFormatString = "<DATE>{0}<TIME>{0}<MSEC>{0}<Ask>{0}<Bid>{0}<BuyCount>{0}<BuySqty>{0}<SellCount>{0}<SellSqty>{0}<TheoriticPrice>{0}<Volatility>{0}<OpenInterest>";

        //static readonly TimeSpan MoscowToLocalTimeShift = new TimeSpan(4, 0, 0);

        static void Main(string[] args)
        {
            if (args.Length < 1)
            {
                string exeName = AppDomain.CurrentDomain.FriendlyName;
                string useage = "Конвертер файлов формата \"fid\"\n" +
                  "\t" + exeName + " [/t, /b] исходный_файл [конечный_файл]\n" +
                  "Если не указан параметр /t или /b, то направление преобразования будет исходя из расширения исходного файла\n" +
                  "Параметры:\n" +
                  "\t /t - преобразовать двоичный файл (.fid) в текстовый (.txt)\n" +
                  "\t /b - преобразовать текстовый файл (.txt) в двоичный (.fid)\n" +
                  "\n" +
                  "Пример:\n\t" + exeName + " SiH2.fid\n" +
                  "\t\t Конвертировать двоичный файл SiH2.fid в текстовый SiH2.txt\n" +
                  "\t" + exeName + " SiH2.txt\n" +
                  "\t\t Конвертировать текстовый файл SiH2.txt в двоичный файл SiH2.fid\n" +
                  "\t" + exeName + " SiH2.txt NewSiH2.fid\n" +
                  "\t\t Конвертировать текстовый файл SiH2.txt в двоичный файл NewSiH2.fid\n" +
                  "\t" + exeName + " /t SiH2.fid SiH2.txt\n" +
                  "\t\t Конвертировать двоичный файл SiH2.fid в текстовый SiH2.txt\n" +
                  "\n";
                Console.WriteLine(useage);
                return;
            }

            int i = 0;
            string key = args[i];
            if (!key.StartsWith("-") && !key.StartsWith("/"))
                key = string.Empty;
            else
            {
                key = key.TrimStart('-', '/').ToUpper();
                i++;
            }

            string srcFile = args[i++];
            if (!File.Exists(srcFile))
            {
                Console.WriteLine(string.Format("Файл {0} не найден.", srcFile));
                return;
            }

            string dstFile = string.Empty;
            if (args.Length > i)
                dstFile = args[i++];

            if (string.IsNullOrEmpty(key))
            {
                string ext = Path.GetExtension(srcFile);
                if (string.Compare(ext, FininfoCache.Extension, StringComparison.InvariantCultureIgnoreCase) == 0)
                    key = "T";
                else if (string.Compare(ext, ".txt", StringComparison.InvariantCultureIgnoreCase) == 0)
                    key = "B";
            }

            try
            {
                if (key.StartsWith("T"))
                    if (dstFile == string.Empty)
                        ConvertToTextFile(srcFile);
                    else
                        ConvertToTextFile(srcFile, dstFile);
                else if (key.StartsWith("B"))
                    if (dstFile == string.Empty)
                        ConvertToBinaryFile(srcFile);
                    else
                        ConvertToBinaryFile(srcFile, dstFile);

            }
            catch (IOException e)
            {
                Console.WriteLine("Не могу прочитать файл: {0}\n{1}", srcFile, e.Message);
                Trace.TraceInformation("{0} Не могу прочитать файл: {1}\n{2}", DateTime.Now, srcFile, e.Message);
            }
            catch (Exception e)
            {
                Console.WriteLine("Не могу разобрать файл: {0}\n{1}", srcFile, e.Message);
                Trace.TraceInformation("{0} Не могу разобрать файл: {1}\n{2}", DateTime.Now, srcFile, e.Message);
            }
        }

        public static void ConvertToTextFile(string sourceFileName)
        {
            ConvertToTextFile(sourceFileName, Path.ChangeExtension(sourceFileName, ".txt"));
        }

        public static void ConvertToTextFile(string sourceFileName, string destFileName)
        {
            using (var fs = File.Open(sourceFileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite))
            {
                FininfoCache.CheckFormat(fs);

                var recCount = (fs.Length - FininfoCache.HeaderSize) / FininfoCache.RecSize;
                if (recCount <= 0)
                    return;

                //var tf = new TimeSpan(0, 5, 0);
                // Получаем строку из форматстроки заголовка. Подствляем разделитель.
                var headerString = string.Format(HeaderFormatString, Consts.Separator);

                using (var ws = File.Open(destFileName, FileMode.Create, FileAccess.Write, FileShare.Read))
                {
                    var br = new BinaryReader(fs);
                    using (var sw = new StreamWriter(ws))
                    {
                        sw.WriteLine(headerString);
                        var cc = Thread.CurrentThread.CurrentCulture;
                        Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
                        try
                        {
                            long i = 0;
                            while (fs.Position < fs.Length)
                            {
                                var fi = FinInfoCacheHelper.ReadFromStream(i++, br);
                                WriteRecordAsText(sw, fi);
                            }
                        }
                        catch (EndOfStreamException) { }
                        finally
                        {
                            Thread.CurrentThread.CurrentCulture = cc;
                        }
                    }
                }
            }
        }

        public static void ConvertToBinaryFile(string sourceFileName)
        {
            ConvertToBinaryFile(sourceFileName, Path.ChangeExtension(sourceFileName, FininfoCache.Extension));
        }

        public static void ConvertToBinaryFile(string sourceFileName, string destFileName)
        {
            using (var srcFile = File.Open(sourceFileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite))
            {
                var sr = new StreamReader(srcFile);
                var line = sr.ReadLine();
                if (string.IsNullOrEmpty(line))
                    return;

                // Получаем строку из форматстроки заголовка. Подствляем разделитель.
                var headerString = string.Format(HeaderFormatString, Consts.Separator);

                line = line.Trim();
                if (!line.Equals(headerString, StringComparison.InvariantCultureIgnoreCase))
                    throw new ArgumentException("Неподходящий формат файла: " + sourceFileName);

                using (var dstFile = File.Open(destFileName, FileMode.Create, FileAccess.ReadWrite, FileShare.Read))
                {
                    dstFile.WriteByte(FininfoCache.FileFormatVersion); // Номер версии формата
                    using (var bw = new BinaryWriter(dstFile))
                    {
                        var cc = Thread.CurrentThread.CurrentCulture;
                        Thread.CurrentThread.CurrentCulture = CultureInfo.InvariantCulture;
                        try
                        {
                            FinInfoBlock fi;
                            while (!sr.EndOfStream)
                            {
                                ReadRecordAsText(sr, out fi);
                                fi.WriteToStream(bw);
                            }
                        }
                        finally
                        {
                            Thread.CurrentThread.CurrentCulture = cc;
                        }
                    }
                }
            }
        }


        static void WriteRecordAsText(StreamWriter sw, FinInfoBlock fi)
        {
            const string comma = ",";

            //var date = fi.LastUpdate.ToLocalTime();
            // Переводим универсальное время в московское. 
            var date = TimeZoneInfo.ConvertTimeFromUtc(fi.LastUpdate,
                                                   TimeZoneInfo.FindSystemTimeZoneById("Russian Standard Time"));
            //date += MoscowToLocalTimeShift; // Дополнительная коррекция часового пояса
            sw.Write(date.ToString("yyyyMMdd"));
            sw.Write(comma);
            sw.Write(date.ToString("HHmmss"));
            sw.Write(comma);
            sw.Write(date.ToString("fff"));
            sw.Write(comma);
            sw.Write(fi.Ask);
            sw.Write(comma);
            sw.Write(fi.Bid);
            sw.Write(comma);
            sw.Write(fi.BuyCount);
            sw.Write(comma);
            sw.Write(fi.BuySqty);
            sw.Write(comma);
            sw.Write(fi.SellCount);
            sw.Write(comma);
            sw.Write(fi.SellSqty);
            sw.Write(comma);
            sw.Write(fi.TheoriticPrice);
            sw.Write(comma);
            sw.Write(fi.Volatility);
            sw.Write(comma);
            sw.Write(fi.OpenInterest);
            sw.WriteLine();
        }

        static void ReadRecordAsText(StreamReader sw, out FinInfoBlock fi)
        {
            const char comma = ',';
            var line = sw.ReadLine();
            if (string.IsNullOrEmpty(line))
                throw new Exception("В исходном файле встретилась пустая строка. Это не допустимо");

            var items = line.Split(comma);
            if (items.Length < 11)
                throw new Exception("В исходном файле встретилась строка состоящая менее чем из 11 элементов. Это не допустимо");

            fi = new FinInfoBlock();
            var ci = CultureInfo.InvariantCulture;
            var d = DateTime.ParseExact(items[0] + items[1] + items[2], "yyyyMMddHHmmssfff", ci);
            //d -= MoscowToLocalTimeShift; // Дополнительная коррекция часового пояса
            fi.LastUpdate = d.ToUniversalTime();
            fi.Ask = double.Parse(items[3], ci);
            fi.Bid = double.Parse(items[4], ci);
            fi.BuyCount = double.Parse(items[5], ci);
            fi.BuySqty = double.Parse(items[6], ci);
            fi.SellCount = double.Parse(items[7], ci);
            fi.SellSqty = double.Parse(items[8], ci);
            fi.TheoriticPrice = double.Parse(items[9], ci);
            fi.Volatility = double.Parse(items[10], ci);
        }
    }
}
